import scipy.constants as sc
import numpy as np
import logging
from importlib.resources import files

logging.basicConfig(level=logging.INFO)

''' This section details the calculation of the detailed balance limit efficiency and associated values.'''

h = sc.h        # Planck's constant (J·s)
c = sc.c        # Speed of light (m/s)
k = sc.k        # Boltzmann constant (J/K)
q = sc.e        # Elementary charge (Coulombs)

# Convert the spectrum to the useful units - taken from https://github.com/kaklin/sq-limit?tab=readme-ov-file

def load_spectrum(spectrum_type):

    if spectrum_type == "AM1.5":
        filename = 'ASTMG173.csv'

    elif spectrum_type == "Fluorescent":
        filename = 'fluorescent.csv'

    elif spectrum_type == "Blue LED":
        filename = 'led_blue.csv'

    elif spectrum_type == "Green LED":
        filename = 'led_green.csv'

    elif spectrum_type == "Red LED":
        filename = 'led_red.csv'

    elif spectrum_type == "White LED":
        filename = 'led_white.csv'

    elif spectrum_type == "IR LED":
        filename = 'led_ir.csv'

    elif spectrum_type == "Photopic":
        filename = 'photopic.csv'

    else:
        print("Unrecognisable spectrum selected")
        print("Options: AM1.5, Fluorescent, Blue LED, Green LED, Red LED, White LED, IR LED, Photopic")
        print("reverting to AM1.5")

        filename = 'ASTMG173.csv'

    csv_path = files("placeholder.resources") / f"{filename}"

    with csv_path.open("r", encoding="utf-8") as f:
        spectrum = np.loadtxt(f, delimiter=",", skiprows=1)

    return spectrum

def convert_spectrum(spectrum):

    """
    Converts the input spectrum from standard format to the required units for this code. 

    Parameters:
        spectrum(numpy.ndarray): Input spectrum loaded from a csv file with numpy.loadtxt.

    Returns:
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.

    Spectrum input:
        y: Irradiance (W/m2/nm)
        x: Wavelength (nm)
    Converted output:
        y: Number of photons (Np/m2/s/dE)
        x: Energy (eV)
    """
    converted = np.copy(spectrum)
    converted[:, 0] = converted[:, 0] * 1e-9  # wavelength to m
    converted[:, 1] = converted[:, 1] / 1e-9  # irradiance to W/m2/m (from W/m2/nm)

    E = h * c / converted[:, 0]
    d_lambda_d_E = h * c / E**2
    converted[:, 1] = converted[:, 1] * d_lambda_d_E * q / E
    converted[:, 0] = E / q

    return converted


def photons_above_bandgap(E_gap, photon_spectrum):
    """Counts number of photons above given bandgap.
    
    Parameters:
        E_gap(float): Optical Band Gap in eV  
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.

    Returns:
       (float): Integration of the spectrum for the number of photons above the bandgap.  
    """
    indexes = np.where(photon_spectrum[:, 0] > E_gap)
    y = photon_spectrum[indexes, 1][0]
    x = photon_spectrum[indexes, 0][0]
    return np.trapz(y[::-1], x[::-1])

def rr0(E_gap, photon_spectrum, Tcell):
    '''
    Calculates the radiative recombination rate at 0 Quasi-Fermi Level splitting. 

    Parameters: 
        E_gap(float):  Optical Band Gap in eV  
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.
        Tcell(float): Operating temperature of the cell in K

    Returns:
        Radiative recomination rate(float) in cm⁻³s⁻¹ 

    '''
    k_eV = k / q
    h_eV = h / q
    const = (2 * np.pi) / (c**2 * h_eV**3)

    k_eV = k / q
    E = photon_spectrum[::-1, ]  # in increasing order of bandgap energy
    egap_index = np.where(E[:, 0] >= E_gap)
    numerator = E[:, 0]**2
    exponential_in = E[:, 0] / (k_eV * Tcell)
    denominator = np.exp(exponential_in) - 1
    integrand = numerator / denominator

    integral = np.trapz(integrand[egap_index], E[egap_index, 0])

    result = const * integral
    return result[0]

def recomb_rate(E_gap, photon_spectrum, voltage, Tcell):
    '''
    Calculates the radiative recombination rate. 

    Parameters: 
        E_gap(float):  Optical Band Gap in eV  
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.
        voltage(float): Open circuit voltage in V
        Tcell(float): Operating temperature of the cell in K

    Returns:
        Radiative recomination rate(float) in cm⁻³s⁻¹ 

    '''

    print ('recomb rate')
    return q * rr0(E_gap, photon_spectrum) * np.exp(q * voltage / (k * Tcell))

def current_density(E_gap, photon_spectrum, voltage, Tcell):
    '''
    Calculates the current density. 

    Parameters: 
        E_gap(float):  Optical Band Gap in eV  
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.
        voltage(float): Open circuit voltage in V
        Tcell(float): Operating temperature of the cell in K

    Returns:
        Current density (float):  Current that flows across a cross sectional area in C cm⁻³s⁻¹. 

    '''

    return q * (photons_above_bandgap(E_gap, photon_spectrum) - rr0(E_gap, photon_spectrum, Tcell) * np.exp(q * voltage / (k * Tcell)))


def jsc(E_gap, photon_spectrum, Tcell):

    '''
    Calculates the current density. 

    Parameters: 
        E_gap(float):  Optical Band Gap in eV  
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.

    Returns:
        Short circuit current density (float):  Current that flows across a cross sectional area at 0 applied voltage in C cm⁻³s⁻¹.  

    '''
    
    return current_density(E_gap, photon_spectrum, 0, Tcell)


def voc(E_gap, photon_spectrum, Tcell):
    '''
    Calculates the open circuit voltage.

    Parameters: 
        E_gap(float):  Optical Band Gap in eV  
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.
        Tcell(float): Operating temperature of the cell in K

    Returns:
        Open circuit voltage (float):  Maximum voltage across a solar cell with no current flow in V. 

    '''

    # print 'voc'
    return (k * Tcell / q) * np.log(photons_above_bandgap(E_gap, photon_spectrum) / rr0(E_gap, photon_spectrum, Tcell))

def v_at_mpp(E_gap, photon_spectrum):

    '''
    Calculates the voltage at maximum power point (mpp) of a solar cell.

    Parameters: 
        E_gap(float):  Optical Band Gap in eV  
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.

    Returns:
        Voltage at MPP (float):  Voltage across a solar cell at the maximum power point in V. 

    '''

    v_open = voc(E_gap, photon_spectrum)
    # print v_open
    v = np.linspace(0, v_open)
    index = np.where(v * current_density(E_gap, photon_spectrum, v)==max(v * current_density(E_gap, photon_spectrum, v)))
    return v[index][0]


def j_at_mpp(E_gap, photon_spectrum):

    '''
    Calculates the current at maximum power point (mpp) of a solar cell.

    Parameters: 
        E_gap(float):  Optical Band Gap in eV  
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.

    Returns:
       Current at MPP (float):  Current across a solar cell at the maximum power point. 

    '''

    return max_power(E_gap, photon_spectrum) / v_at_mpp(E_gap, photon_spectrum)


def max_power(E_gap, photon_spectrum, Tcell):

    '''
    Calculates the maximum power of a solar cell.

    Parameters: 
        E_gap(float):  Optical Band Gap in eV  
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.
        Tcell(float): Operating temperature of the cell in K

    Returns:
       Maximum power (float):  Maximum power of the solar cell in V C cm⁻³ s⁻¹. 

    '''

    v_open = voc(E_gap, photon_spectrum, Tcell)
    v = np.linspace(0, v_open)
    index = np.where(v * current_density(E_gap, photon_spectrum, v, Tcell)==max(v * current_density(E_gap, photon_spectrum, v, Tcell)))
    return max(v * current_density(E_gap, photon_spectrum, v, Tcell))


def max_eff(E_gap, photon_spectrum, Tcell):

    '''
    Calculates the maximum efficiency of a solar cell.

    Parameters: 
        E_gap(float):  Optical Band Gap in eV  
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.
        Tcell(float): Operating temperature of the cell in K

    Returns:
       Maximum efficiency (float):  Maximum effeciency of the solar cell relative to the total irradiance in %.
    '''

    photon_spectrum_1 = photon_spectrum[::-1, 1] 
    photon_spectrum_0 = photon_spectrum[::-1, 0]

    irradiance =  np.trapz(photon_spectrum_1 * q * photon_spectrum_0, photon_spectrum_0)
    return max_power(E_gap, photon_spectrum, Tcell) / irradiance

def FillFactor(E_gap, photon_spectrum, Tcell):

    '''
    Calculates the fill factor of a solar cell.

    Parameters: 
        E_gap(float):  Optical Band Gap in eV  
        photon_spectrum(numpy.ndarray): Output spectrum as numpy ndarray.
        Tcell(float): Operating temperature of the cell in K

    Returns:
       fill_factor (float):  The fill factor of a solar cell.
    '''

    j_sc = jsc(E_gap, photon_spectrum)
    v_oc = voc(E_gap, photon_spectrum, Tcell)
    v_mpp = v_at_mpp(E_gap, photon_spectrum)
    j_mpp = j_at_mpp(E_gap, photon_spectrum)

    fill_factor = (j_mpp * v_mpp) / (j_sc, v_oc)

    return fill_factor





from typing import Dict, List, Optional, Union
from importlib.resources import files
import json 
import solphin.resources

from monty.serialization import loadfn
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.sets import VaspInputSet
from pymatgen.io.vasp import Kpoints

def read_structure_pmg(filename):
    """
    Reads a structure file (POSCAR/CONTCAR, CIF, XYZ, etc.) into a pymatgen Structure object.
    
    Parameters
    ----------
    filename : str
        Path to the structure file. Supported formats include:
        - VASP POSCAR/CONTCAR
        - CIF (.cif)
        - XYZ (.xyz)
        - JSON (.json)
        - Other pymatgen-supported formats
    
    Returns
    -------
    Structure
        A pymatgen Structure object with lattice, species, and atomic positions.
    """
    structure = Structure.from_file(filename)
    return structure

def _load_config(fname: str) -> Dict[str, Dict[str, Union[str, Dict[str, str]]]]:
    """
    Load configuration information from a JSON file.

    Args:
        fname (str): The name of the JSON file to load.

    Returns:
        dict: A dictionary containing the configuration information.
    """
    resource_path = files("placeholder.resources") / fname  # adjust module name
    with resource_path.open("r", encoding="utf-8") as f:
        config = json.load(f)
    return config

def determine_potcar_functional(
    recipe: str,
    potcar_functional: Optional[str],
    config: Dict[str, Dict[str, Union[str, Dict[str, str]]]],
) -> str:
    """
    Determine the POTCAR functional to use based on the recipe.

    Args:
        recipe (str): The name of the recipe.
        potcar_functional (str, optional): The provided POTCAR functional.
        config (dict): The configuration dictionary.

    Returns:
        str: The determined POTCAR functional.
    """
    if potcar_functional:
        return potcar_functional
    if recipe == "LDA":
        return "LDA_64"
    return config["POTCAR_FUNCTIONAL"]


def prepare_incar(
    recipe: str,
    patches: List[str],
    config: Dict[str, Dict[str, Union[str, Dict[str, str]]]],
) -> Dict[str, Union[str, int, float]]:
    """
    Prepare the INCAR settings for the VASP calculation.

    Args:
        recipe (str): The name of the recipe.
        patches (list): A list of patches to apply.
        config (dict): The configuration dictionary.

    Returns:
        dict: The prepared INCAR settings.
    """

    incar = config["INCAR"][recipe]
    if recipe in ["HSE06", "PBE0"]:
        incar.update({"NCORE": 4})

    for patch in patches:
        if patch != "gamma_only":
            if patch == "defect":
                incar.update(config["PATCHES"]["relax_atoms"])
                incar.update(config["PATCHES"]["spin_polarised"])
            else:
                incar.update(config["PATCHES"][patch])
    return incar

def create_vasp_set(
    structure: Structure,
    incar: Dict[str, Union[str, int, float]],
    potcar_functional: str,
    config: Dict[str, Dict[str, Union[str, Dict[str, str]]]],
    **calc_kwargs,
) -> VaspInputSet:
    """
    Create the VASP set for the calculation.

    Args:
        structure (Structure): The structure to calculate.
        incar (dict): The INCAR settings.
        potcar_functional (str): The POTCAR functional to use.
        config (dict): The configuration dictionary.

    Returns:
        DictSet: The created VASP set.
    """
    return VaspInputSet(
        structure,
        {
            "INCAR": incar,
            "POTCAR": config["POTCAR"],
            "POTCAR_FUNCTIONAL": potcar_functional,
        },
        **calc_kwargs,
    )

def prepare_vdw_tags(recipe: str, patches: List[str]) -> Dict[str, Union[int, float]]:
    """
    Prepare the VDW tags for the INCAR settings.

    Args:
        recipe (str): The name of the recipe.
        patches (list): A list of patches to apply.

    Returns:
        dict: The prepared VDW tags.
    """
    if "vdw_d3_bj" in patches:
        if recipe == "HSE06":
            return {
                "IVDW": 12,
                "VDW_S8": 2.310,
                "VDW_A1": 0.383,
                "VDW_A2": 5.685,
            }
        return {"IVDW": 12}
    if "vdw_d3" in patches:
        if recipe == "HSE06":
            return {"IVDW": 11, "VDW_S8": 0.109, "VDW_SR": 1.129}
        return {"IVDW": 11}
    if "vdw_d4" in patches:
        if recipe == "HSE06":
            return {"IVDW": 13, 
                    "VDW_S8" : 1.19528249, 
                    "VDW_A1" : 0.38663183, 
                    "VDW_A2" : 5.19133469}
        return {"IVDW": 13}
    if "rvv10" in patches:
        if recipe == "R2SCAN":
            return {"LUSE_VDW" : True,
                    "BPARAM": 11.95,
                    "CPARAM": 0.0093}
    return {}

def apply_patches(
    vasp_set: VaspInputSet,
    patches: List[str],
    recipe: str,
    incar: Dict[str, Union[str, int, float]],
) -> None:
    """
    Apply patches to the VASP set.

    Args:
        vasp_set (DictSet): The VASP set.
        patches (list): A list of patches to apply.
        recipe (str): The name of the recipe.
        incar (dict): The INCAR settings.
    """
    if "relax_cell" in patches:
        encut = vasp_set.user_incar_settings.get("ENCUT", incar["ENCUT"]) * 1.3
        vasp_set.user_incar_settings["ENCUT"] = encut

    if "gamma_only" in patches:
        vasp_set.user_kpoints_settings = Kpoints(kpts=((1, 1, 1)))

    if "vdw_d3_bj" in patches or "vdw_d3" in patches or "vdw_d4":
        vdw_tags = prepare_vdw_tags(recipe, patches)
        vasp_set.user_incar_settings.update(vdw_tags)

    if "lobster" in patches and "NBANDS" not in vasp_set.user_incar_settings:
        vasp_set.user_incar_settings["NBANDS"] = vasp_set.estimate_nbands() * 2


def write_vasp_calculation(
    structure: Structure,
    recipe: str,
    out_dir: str,
    patches: Optional[List[str]] = None,
    potcar_functional: Optional[str] = None,
    **calc_kwargs,
) -> Optional[Dict[str, int]]:
    """
    Write a VASP calculation from a recipe.

    Args:
        structure (pymatgen.core.structure.Structure): The structure to calculate.
        recipe (str): The name of the recipe to use.
        out_dir (str): The output directory for the calculation.
        patches (list, optional): A list of patches to apply to the recipe. Defaults to None.
        get_stats (bool, optional): Whether to return statistics on the calculation. Defaults to False.
        potcar_functional (str, optional): The POTCAR functional to use. Defaults to None.

    Returns:
        dict: A dictionary containing the estimated number of bands and the number of electrons in the system, if get_stats is True.
    """
    patches = patches or []

    config = _load_config("base_recipes.json")
    potcar_functional = determine_potcar_functional(recipe, potcar_functional, config)
    incar = prepare_incar(recipe, patches, config)
    vasp_set = create_vasp_set(
        structure, incar, potcar_functional, config, **calc_kwargs
    )

    apply_patches(vasp_set, patches, recipe, incar)
    vasp_set.write_input(out_dir)





import solphin.db_fom as db_fom
import numpy as np
import matplotlib.pyplot as plt

import logging

logging.basicConfig(level=logging.INFO)

def photons_above_bandgap_plot(spectrum, Egap):
    """Plot of photons above bandgap as a function of bandgap"""
    a = np.copy(spectrum)
    for row in a:
        # print row
        row[1] = db_fom.photons_above_bandgap(row[0], spectrum)
    plt.plot(a[:, 0], a[:, 1], color='#231123')

    p_above_1_1 = db_fom.photons_above_bandgap(Egap, spectrum)
    plt.plot([Egap], [p_above_1_1], 'ro', color='#FF6666')
    plt.text(Egap+0.05, p_above_1_1, '{}eV, {:.4}'.format(Egap, p_above_1_1))

    plt.xlabel('$E_{gap}$ (eV)')
    plt.ylabel('# Photons $m^{-2}s^{-1}$')
    plt.title('Number of above-bandgap \nphotons as a function of bandgap')
    plt.show()

def iv_curve_plot(egap, spectrum, Tcell, power=False):
    """Plots the ideal IV curve, or the ideal power for a given material"""
    v_open = db_fom.voc(egap, spectrum, Tcell)
    v = np.linspace(0, v_open)
    if power:
        p =  v * db_fom.current_density(egap, spectrum, v, Tcell)
        plt.xlabel('Voltage (V)')
        plt.ylabel('Power generated ($W$)')
        plt.title('Power Curve')
        plt.plot(v, p, color='#231123')
    else:
        i =  db_fom.current_density(egap, spectrum, v, Tcell)
        plt.xlabel('Voltage (V)')
        plt.ylabel('Current density $J$ ($Am^{-2}$)')
        plt.title('IV Curve')
        plt.plot(v, i, color='#231123')


def iv_curve_plot_2(egap, spectrum, Tcell, power=False):
    """Plots the ideal IV curve, and the ideal power for a given material"""
    v_open = db_fom.voc(egap, spectrum, Tcell)
    v = np.linspace(0, v_open)

    fig, ax1 = plt.subplots()
    p =  v * db_fom.current_density(egap, spectrum, v, Tcell)
    i =  db_fom.current_density(egap, spectrum, v, Tcell)
    
    ax1.plot(v, i, color='#231123')
    ax1.set_xlabel('Voltage (V)')
    ax1.set_ylabel('Current density $J$ ($Am^{-2}$)')
    ax1.legend(['Current'], loc=2)
    
    ax2 = ax1.twinx()
    ax2.plot(v, p, color='#FF6666')
    ax2.set_ylabel('Power generated ($W$)')
    ax2.legend(['Power'], loc=3)
    return


def sq_limit_plot(spectrum, Egap, Tcell):
    # Plot the famous SQ limit
    a = np.copy(spectrum)
    # Not for whole array hack to remove divide by 0 errors
    for row in a[2:]:
        # print row
        row[1] = db_fom.max_eff(row[0], spectrum, Tcell)
    # Not plotting whole array becase some bad values happen
    plt.plot(a[2:, 0], a[2:, 1])
    e_gap = Egap
    p_above_1_1 = db_fom.max_eff(e_gap, spectrum, Tcell)
    plt.plot([e_gap], [p_above_1_1], 'ro')
    plt.text(e_gap+0.05, p_above_1_1, '{}eV, {:.4}'.format(e_gap, p_above_1_1))

    plt.xlabel('$E_{gap}$ (eV)')
    plt.ylabel('Max efficiency')
    plt.title('SQ Limit')


import numpy as np
import logging

logging.basicConfig(level=logging.INFO)

'''
This module calculates all the components required for the Γₚᵥ Figure of Merit from Andrea Crovetto 2024 J. Phys. Energy 6 025009
'''

def Final_equation(E_gap, alpha, tau, sigma, dos_mass, dop_density, epsilon, mu):

    ''' Calculates the total Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        sigma(float): Spectral dispersion of the absorption coefficient spectrum, unitless
        dos_mass(float): Density of States effective mass in m₀
        dop_density(float): Doping density in cm⁻³
        epsilon(float): Static dielectric constant, unitless
        mu(float): Charge carrier mobility in cm²V⁻¹s⁻¹

    Returns:
        PV_FOM(float): Γₚᵥ Photovoltaic Figure of Merit from Crovetto 2024, unitless
    '''

    E_gap_2_5 = E_gap ** 2.5
    E_gap_0_8 = E_gap ** -0.8

    D_denominator = Final_D_denominator(E_gap, alpha, tau, dop_density, epsilon)
    T_denominator = Final_T_denominator(E_gap, alpha, tau, sigma, dos_mass, dop_density, epsilon, mu)
    S_denominator = Final_S_denominator(E_gap, alpha, tau, dos_mass, dop_density, mu)

    numerator = Final_numerator(E_gap, alpha, tau, sigma, dos_mass, dop_density, epsilon)

    denominator = D_denominator * T_denominator * S_denominator

    PV_FOM = E_gap_2_5 * ((numerator / denominator) ** E_gap_0_8)

    return PV_FOM

def Final_numerator(E_gap, alpha, tau, sigma, dos_mass, dop_density, epsilon):

    ''' Calculates the numerator for the Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        sigma(float): Spectral dispersion of the absorption coefficient spectrum, unitless
        dos_mass(float): Density of States effective mass in m₀
        dop_density(float): Doping density in cm⁻³
        epsilon(float): Static dielectric constant, unitless

    Returns:
        numerator(float): Γₚᵥ numerator only from Crovetto 2024, unitless
    '''

    A_1 = A_1_equation(E_gap, alpha, tau, sigma, dos_mass)
    A_2 = A_2_equation(alpha, tau, sigma)
    D_1 = D_1_equation(alpha, dop_density, epsilon)

    numerator = A_1 * A_2 * D_1

    return numerator

def Final_D_denominator(E_gap, alpha, tau, dop_density, epsilon):

    ''' Calculates the D₂D₃D₄ component of the denominator of Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        dop_density(float): Doping density in cm⁻³
        epsilon(float): Static dielectric constant, unitless

    Returns:
        D_denominator(float): The D₂D₃D₄ component of Γₚᵥ from Crovetto 2024, unitless
    '''

    D_2 = D_2_equation(E_gap, alpha, tau, dop_density)
    D_3 = D_3_equation(E_gap, alpha, tau, dop_density, epsilon)
    D_4 = D_4_equation(E_gap, alpha, tau)

    D_denominator = D_2 * D_3 * D_4

    return D_denominator

def Final_T_denominator(E_gap, alpha, tau, sigma, dos_mass, dop_density, epsilon, mu):
    
    ''' Calculates the 1 + (T₁T₂T₃) component of the Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        sigma(float): Spectral dispersion of the absorption coefficient spectrum, unitless
        dos_mass(float): Density of States effective mass in m₀
        dop_density(float): Doping density in cm⁻³
        epsilon(float): Static dielectric constant, unitless
        mu(float): Charge carrier mobility in cm²V⁻¹s⁻¹

    Returns:
        T_denominator(float): The 1 + (T₁T₂T₃) component of Γₚᵥ from Crovetto 2024, unitless
    '''
    
    T_1 = T_1_equation(E_gap, dos_mass, epsilon, mu)
    T_2 = T_2_equation(E_gap, tau, sigma, dop_density)
    T_3 = T_3_equation(E_gap, alpha, dos_mass, dop_density)

    T_denominator = 1 + (T_1 * T_2 * T_3)

    return T_denominator

def Final_S_denominator(E_gap, alpha, tau, dos_mass, dop_density, mu):

    ''' Calculates the 1 + (S₁S₂) component of the Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        dos_mass(float): Density of States effective mass in m₀
        dop_density(float): Doping density in cm⁻³
        mu(float): Charge carrier mobility in cm²V⁻¹s⁻¹

    Returns:
        S_denominator(float): The 1 + (S₁S₂) component of Γₚᵥ from Crovetto 2024, unitless
    '''

    S_1 = S_1_equation(E_gap, alpha, tau, dos_mass, mu)
    S_2 = S_2_equation(alpha, dop_density, mu)

    S_denominator = 1 + (S_1 * S_2)

    return S_denominator


def A_1_equation(E_gap, alpha, tau, sigma, dos_mass):

    ''' Calculates the A₁ component of Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        sigma(float): Spectral dispersion of the absorption coefficient spectrum, unitless
        dos_mass(float): Density of States effective mass in m₀

    Returns:
        A_1(float): The A₁ component of Γₚᵥ from Crovetto 2024, unitless
    '''

    a_1 = 0.295
    a_2 = 0.185

    E_gap_0_5 = E_gap ** 0.5

    power_1 = - a_2 * E_gap_0_5
    power_2 = sigma ** power_1

    numerator = a_1 * tau * (alpha ** power_2)
    denominator = dos_mass ** 2

    A_1 = numerator/denominator

    return A_1

def A_2_equation(alpha, tau, sigma):

    ''' Calculates the A₂ component of Γₚᵥ from Crovetto 2024

    Parameters:
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        sigma(float): Spectral dispersion of the absorption coefficient spectrum, unitless
       
    Returns:
        A_2(float): The A₂ component of Γₚᵥ from Crovetto 2024, unitless
    '''

    a_3 = 1.0e-7

    fraction = (sigma ** 10) / (alpha * tau)

    A_2 = 1 + ((a_3 * fraction) ** 0.4)

    return A_2

# D_1 Equation 

def D_1_equation(alpha, dop_density, epsilon):

    ''' Calculates the D₁ component of Γₚᵥ from Crovetto 2024

    Parameters:
        alpha(float): Spectral average of incident light in cm⁻¹
        dop_density(float): Doping density in cm⁻³
        epsilon(float): Static dielectric constant, unitless

    Returns:
        D_1(float): The D₁ component of Γₚᵥ from Crovetto 2024, unitless
    '''

    d_1 = 4.4e-5
    d_2 = 39

    log_bracket = alpha / d_2

    power = 0.22 * np.log10(log_bracket)

    denominator = (epsilon ** 0.8) * (alpha ** 2)

    D_1 = (1 + d_1 * (dop_density / denominator)) ** power

    return D_1

# D_2 Equation

def D_2_equation(E_gap, alpha, tau, dop_density):

    ''' Calculates the D₂ component of Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        dop_density(float): Doping density in cm⁻³

    Returns:
        D_2(float): The D₂ component of Γₚᵥ from Crovetto 2024, unitless
    '''

    d_3 = 1e-21

    E_gap_4 = E_gap ** 4
    power = 0.05 * E_gap_4

    fraction = dop_density / ((alpha ** 2) * tau)

    D_2 = (1 + d_3 * fraction) ** power

    return D_2

# D_3 Equation 

def D_3_equation(E_gap, alpha, tau, dop_density, epsilon): 

    ''' Calculates the D₃ component of Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        dop_density(float): Doping density in cm⁻³
        epsilon(float): Static dielectric constant, unitless

    Returns:
        D_3(float): The D₃ component of Γₚᵥ from Crovetto 2024, unitless
    '''

    d_4 = 2.1e4
    d_5 = 50

    E_gap_8_5 = E_gap ** 8.5
    E_gap_1_5 = E_gap ** -1.5

    power_num_1 = 0.68 * E_gap_1_5
    power_num_2 = np.log10((10 * dop_density) / epsilon) / d_5
    power_denum = (E_gap - 1.5) / 0.1

    numerator = (d_4 * E_gap_8_5 * tau * (alpha ** power_num_1)) ** power_num_2
    denominator = 1 + (10 ** power_denum)

    D_3 = 1 + (numerator / denominator)

    return D_3

# D_4 Equation 

def D_4_equation(E_gap, alpha, tau):

    ''' Calculates the D₄ component of Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
       
    Returns:
        D_4(float): The D₄ component of Γₚᵥ from Crovetto 2024, unitless
    '''

    d_6 = 7.7e-7

    fraction = d_6 / ((E_gap ** 17) * alpha * tau)

    D_4 = 1 + (fraction ** 0.6)

    return D_4 

# T_1 Equation

def T_1_equation(E_gap, dos_mass, epsilon, mu):

    ''' Calculates the total Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        dos_mass(float): Density of States effective mass in m₀
        epsilon(float): Static dielectric constant, unitless
        mu(float): Charge carrier mobility in cm²V⁻¹s⁻¹

    Returns:
        PV_FOM(float): Γₚᵥ Photovoltaic Figure of Merit from Crovetto 2024, unitless
    '''

    t_1 = 5.1e-2
    t_2 = 4.6e-2
    E_gap_4_3 = E_gap ** 4.3

    numerator = t_1 * ((E_gap + 0.5) ** 11)
    power = (t_2 * E_gap_4_3) + 0.9
    denominator = dos_mass * (epsilon ** 0.5) * (mu ** power)

    T_1 = numerator / denominator

    return T_1

# T_2 Equation

def T_2_equation(E_gap, tau, sigma, dop_density):

    ''' Calculates the T₂ component of Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        tau(float): Non-radiative recombination lifetime in s
        sigma(float): Spectral dispersion of the absorption coefficient spectrum, unitless
        dop_density(float): Doping density in cm⁻³

    Returns:
        T_2(float): The T₂ component of Γₚᵥ from Crovetto 2024, unitless
    '''

    t_8 = 9.5e-18
    t_9 = 7.80e7

    E_gap_1_25 = E_gap ** 1.25
    E_gap_5_1 = E_gap ** 5.1

    power_main = 0.47 * E_gap_1_25
    power_denom = 0.5 / sigma

    numerator = 1 + t_8 * dop_density * (1 + E_gap_5_1)
    denominator = t_9 * tau * np.power(10, power_denom)

    T_2 = (numerator / denominator) ** power_main

    return T_2

# T_3 Equation 

def T_3_prime_equation(E_gap, alpha):

    ''' Calculates the T₃' component of T₃ used in Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹

    Returns:
        T_3_prime(float): The T₃' component of T₃ from Crovetto 2024, unitless
    '''

    t_3 = 1.9e4
    t_4 = 9.5e-3
    t_5 = 2.4e-4

    E_gap_10 = E_gap ** 10 

    pre_exp = t_3 * (1 + t_4 * E_gap_10)
    power = 0.5 / (1 + t_5 * E_gap_10)
    exponent = np.exp(-0.1 * (alpha ** power))

    T_3_prime = pre_exp * exponent

    return T_3_prime

def T_3_double_prime_equation(E_gap, alpha, dos_mass, dop_density):

    ''' Calculates the T₃'' component of T₃ used in the Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        dos_mass(float): Density of States effective mass in m₀
        dop_density(float): Doping density in cm⁻³

    Returns:
        T_3_double_prime(float): The T₃'' component of T₃ from Crovetto 2024, unitless
    '''

    t_6 = 1.5e5

    power_num = 1 - 0.74 * np.exp(- alpha / t_6)
    power_dom = (E_gap - 1.5) / 0.01

    numerator = (0.16 / dos_mass ** 3) * (((dop_density * (dos_mass ** 3))/0.16) ** power_num) 
    denominator = 1 + 10 ** power_dom

    T_3_double_prime = 1 + numerator / denominator

    return T_3_double_prime


def T_3_equation(E_gap, alpha, dos_mass, dop_density):

    ''' Calculates the T₃ component of Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        dos_mass(float): Density of States effective mass in m₀
        dop_density(float): Doping density in cm⁻³

    Returns:
        T_3(float): The T₃ component of Γₚᵥ from Crovetto 2024, unitless
    '''

    t_7 = 1.6e-3
    E_gap_8 = E_gap ** 8 

    power = (t_7 * E_gap_8) + 0.6

    T_3_prime = T_3_prime_equation(E_gap, alpha)
    T_3_double_prime = T_3_double_prime_equation(E_gap, alpha, dos_mass, dop_density)

    T_3 = (1 + T_3_prime * T_3_double_prime) ** power

    return T_3

def S_1_equation(E_gap, alpha, tau, dos_mass, mu):

    ''' Calculates the S₁ component of Γₚᵥ from Crovetto 2024

    Parameters:
        E_gap(float): Optical Band Gap in eV  
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        dos_mass(float): Density of States effective mass in m₀
        mu(float): Charge carrier mobility in cm²V⁻¹s⁻¹

    Returns:
        S_1(float): The S₁ component of Γₚᵥ from Crovetto 2024, unitless
    '''

    s_1 = 2.4e4
    s_2 = 4e-4

    E_gap_10 = E_gap ** 10

    power_num = 2.4 * E_gap
    power_den = 1 / (1 + s_2 * E_gap_10)

    numerator = ((10 ** power_num) * (alpha ** 0.75) * tau)
    denominator = s_1 * (dos_mass ** 2) * (mu ** power_den)

    S_1 = numerator / denominator

    return S_1   

def S_2_equation(alpha, dop_density, mu):

    ''' Calculates the S₂ component of Γₚᵥ from Crovetto 2024

    Parameters:
        alpha(float): Spectral average of incident light in cm⁻¹
        dop_density(float): Doping density in cm⁻³
        mu(float): Charge carrier mobility in cm²V⁻¹s⁻¹

    Returns:
        S_2(float): The S₂ component of Γₚᵥ from Crovetto 2024, unitless
    '''

    s_3 = 4.8e3

    bracket = (s_3 / alpha) ** 20 

    S_2 = 1 + (bracket * (mu ** 0.5) * np.log10(dop_density))

    return S_2


from pymatgen.io.vasp import Vasprun
import numpy as np
from os.path import join
import scipy.special as sc
import pandas as pd

q=1.60217662E-19
kT=0.0258519975 # eV for T=300K
k=0.000086173325 #eV/K
h=4.135667E-15 #eVs
c=2.9979E+8 #m/s

def calc_dielectric(filename):

    '''Calculates the dielectric constants from a vasprun.xml
    
    Parameters:
        filename(string): filename/ path of the vasprun, typically vasprun.xml.
        
    Returns:
        eps_full(np.array): static dielectric constant (complex, contains both real and imaginary components)
        energies(np.array): energy of the incident radiation eV
        '''

    load_vasprun = Vasprun(filename)
    dielectric = load_vasprun.dielectric

    energies = np.array(dielectric[0])

    real_eps = np.array(dielectric[1])[:, [[0, 3, 5], [3, 1, 4], [5, 4, 2]]]
    imag_eps = np.array(dielectric[2])[:, [[0, 3, 5], [3, 1, 4], [5, 4, 2]]]
    eps_full = real_eps + 1j * imag_eps

    return eps_full, energies

def calc_absorption(eps_full, energies):

    '''Calculates the averages of the real and imaginary components of the refractive index, absorption, losses, real and imaginary components of the 
    static dielectric constant.
    
    Parameters:
        eps_full(np.array): static dielectric constant (complex, contains both real and imaginary components)
        energies(np.array): energy of the incident radiation eV
        
    Returns:
        data(dictionary):
        '''

    # take sqrt of eps matrix; if eps = V S V^-1; then eps^1/2 = V S^{1/2} V^-1;
    eigvals, eigvecs = np.linalg.eig(eps_full)

    # fancy einsum to calculate V S^{1/2} V^-1 at every energy
    n = np.einsum("ijk,ik,ikl->ijl", eigvecs, np.sqrt(eigvals), np.linalg.inv(eigvecs))

    # calculate optical absorption
    alpha = n.imag * energies[:, None, None] * 4 * np.pi / 1.23984212e-4

    # Invert epsilon to obtain energy-loss function
    loss = -np.linalg.inv(eps_full).imag

    eps = np.linalg.eigvals(eps_full).mean(axis=1)
    n = np.linalg.eigvals(n).mean(axis=1)
    loss = np.linalg.eigvalsh(loss).mean(axis=1)
    alpha = np.linalg.eigvalsh(alpha).mean(axis=1)

    data = {
        "eps_real": eps.real,
        "eps_imag": eps.imag,
        "n_real": n.real,
        "n_imag": n.imag,
        "loss": loss,
        "absorption": alpha,
    }

    return data

def print_n_real_file(data, energies):

    filename = 'n_real.dat'
    directory = 'dos'

    if directory:
            filename = join(directory, filename)

    header = "energy(eV)"

    header += " alpha"
    data = np.stack((energies, data['n_real']), axis=1)

    np.savetxt(filename, data, header=header)

def generate_n_real(filename):
     
     eps_full, energies = calc_dielectric(filename)

     data = calc_absorption(eps_full, energies)

     print_n_real_file(data, energies)


def blank_flat(alpha, n, length): 

    #Absorptance for flat scatterer, from Matlab script
    the_c = np.arcsin(1/n[0])
    theta = np.linspace(0.0, the_c, num=200)
    a_len = len(alpha)
    t_len = len(theta)
    toft = np.zeros(a_len*t_len)
    toft = toft.reshape(a_len, t_len)
    absorb_tr = np.zeros(a_len)

    for i, a_pt in enumerate(alpha):
        for j, t_pt in enumerate(theta):
            toft_pt = np.exp(((2*np.multiply((-a_pt), length))/np.cos(t_pt)))
            toft[i, j] = toft_pt
        a_t_1 = np.trapz((toft[i,:]*np.cos(theta)*np.sin(theta)), theta)
        a_t_2 = np.trapz((np.cos(theta)*np.sin(theta)), theta)
        absorb_tr[i] = 1 - (a_t_1/a_t_2)

    absorb = absorb_tr.conjugate()
    return(absorb)


def blank_lambert(alpha, n, length):

    #Absorptance for Lambertian scatterer coating, from Matlab script
    x = np.multiply(2,np.multiply(alpha,length))
    T = np.exp((-x)) - np.multiply(x, np.exp((-x))) + (x**2*sc.exp1(x))
    R = 1.0/(np.multiply(n,n))
    Abs = (1-T)/(1-T+(R*T))
    Abs = np.nan_to_num(Abs)
    Emi = (R*T)/(1-T+(R*T))

    return(Abs)

def blank_eta(spectrum, E, alpha, n, length, Qi, trap):
    #For given scatterer, calculates Blank et al. eta
    dE = E[1]-E[0]
    if trap == 1:
        Abs = blank_flat(alpha, n, length)
    elif trap == 2:
        Abs = blank_lambert(alpha, n, length)
    
    #np divide necessary to have array divide here?
    phibb = 2*np.divide(np.divide(np.multiply(E,E),((h**3)*(c**2))),(np.exp(E/kT)-1))
    phibb = np.nan_to_num(phibb) # NaNs to 0, as in Matlab

    ps_E = spectrum[:, 0]
    phi_sun = spectrum[:, 1]
    
    # works for all E values above 0.03? won't extrapolate for 0
    # values < gap shouldn't be relevant?
    phisun = 10000*np.interp(E, ps_E, phi_sun)
        
    Jsc = q*np.sum(Abs*phisun)*dE
    J0rad = q*np.sum(Abs*phibb)*dE
    
    Rrad = 4*np.pi*np.sum(alpha*(n**2)*phibb)*dE
    Rnrad = (Rrad-Qi*Rrad)/Qi
    
    pe = J0rad/(q*Rrad*length)
    J0 = q*length*(Rnrad + pe*Rrad)
    #Looks like this scans over only voltages between 0 and 2 V
    #need testing for band gaps>2 eV?
    V = np.linspace(0, 2, 1001)
    Pmax = np.max(V*(Jsc - J0*(np.exp(V/kT)-1)))
    eta = Pmax/1000

    return(eta)

def blank_parse(folder):
    
    # Parses outputs from current directory

    abs_data = pd.read_table(f'{folder}/absorption.dat', delim_whitespace=True,
                                skiprows=1, header=None)
    n_data = pd.read_table(f'{folder}/n_real.dat', delim_whitespace=True,
                    skiprows=1, header=None)
    
    E_p = list(abs_data[0])
    alpha_p = list(abs_data[1])
    n_p = list(n_data[1])

    return{"E": E_p, "alpha": alpha_p, "n": n_p}

def blank_calculate(spectrum, folder):

    data = blank_parse(folder)

    E = np.asarray(data["E"])
    alpha = np.asarray(data["alpha"])
    alpha = np.multiply(alpha, 100)
    n = np.asarray(data["n"])

    #Remove data for E>5eV, necessary for speed! Also done in Matlab

    E = np.asarray([o for o in E if o <= 5])
    alpha = alpha[0:(len(E))]
    n = n[0:(len(E))]

    #main, looping over lengths and Qi, outputs eta table
    length_arr = np.logspace(-8.0, -3.0, num=36)
    Qi_arr = np.logspace(0, -6, num=4)
    trap = [1, 2]

    for tr_pt in trap:
        eta_arr = np.zeros(len(length_arr)*(len(Qi_arr)+1))
        eta_arr = eta_arr.reshape(len(length_arr), (len(Qi_arr)+1))
        for k, l_pt in enumerate(length_arr):
            for l, q_pt in enumerate(Qi_arr):
                eta_max = blank_eta(spectrum, E, alpha, n, l_pt, q_pt, tr_pt)
                eta_arr[k, 0] = l_pt
                eta_arr[k, l+1] = eta_max
        if tr_pt == 1:
            head="Thickness[m] \t Eta as fraction for Flat scatterer with Qi = 1.0, 0.01, 1E-4, 1E-6"
            np.savetxt('flat_eta_out', eta_arr, header=head)
        elif tr_pt == 2:
            head="Thickness[m] \t Eta as fraction for Lambertian scatterer with Qi = 1.0, 0.01, 1E-4, 1E-6"
            np.savetxt('lamb_eta_out', eta_arr, header=head)
        


import solphin.pv_fom
import solphin.db_fom
import solphin.final_results
import solphin.db_plotting
import solphin.vasp_inputs
import solphin.spectral
import solphin.optics
from .version import __version__
import solphin.resources


from solphin.pv_fom import Final_equation
from solphin.db_fom import max_eff

# Calculating equation 33 from the FOM paper

def Crovetto_efficiency(E_gap, photon_spectrum, alpha, tau, sigma, dos_mass, dop_density, epsilon, mu, Tcell):

    ''' Calculates the final value for the photovoltaic figure of merit from Crovetto 2024
    
    Parameters:
        E_gap(float): Optical Band Gap in eV  
        photon_spectrum(numpy.ndarrray): Converted input spectrum from DB_FOM.convert_spectrum y: Number of photons (Np/m2/s/dE) x: Energy (eV)
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        sigma(float): Spectral dispersion of the absorption coefficient spectrum, unitless
        dos_mass(float): Density of States effective mass in m₀
        dop_density(float): Doping density in cm⁻³
        epsilon(float): Static dielectric constant, unitless
        mu(float): Charge carrier mobility in cm²V⁻¹s⁻¹
        Tcell(float): Operating temperature of the cell in K

    Returns:
        efficiency(float): Percentage photovoltaic figure of merit efficiency.
    '''

    PV_FOM = Final_equation(E_gap, alpha, tau, sigma, dos_mass, dop_density, epsilon, mu)

    k_1 = 3.3e-1
    k_2 = 9.06e-2
    k_3 = 2.48e-3

    FOM_Pv_235 = PV_FOM ** (-0.235)
    FOM_Pv_869 = PV_FOM ** 0.869
    FOM_Pv_362 = PV_FOM ** (-0.362)

    fraction = (k_1 * FOM_Pv_235) / (1 + k_2 * FOM_Pv_869)
    denom_bracket = 1 + (k_3 * FOM_Pv_362)

    SQ_eff = max_eff(E_gap, photon_spectrum, Tcell)
    SQ = SQ_eff * 100

    efficiency = SQ / ((1 + fraction) * denom_bracket)

    return efficiency

# Crovetto efficiency realtive to SQ limit

def SQ_relative_Crovetto_efficiency(E_gap, photon_spectrum, alpha, tau, sigma, dos_mass, dop_density, epsilon, mu, Tcell):

    ''' Calculates the final value for the photovoltaic figure of merit relative to the SQ limit from Crovetto 2024
    
    Parameters:
        E_gap(float): Optical Band Gap in eV  
        photon_spectrum(numpy.ndarrray): Converted input spectrum from DB_FOM.convert_spectrum y: Number of photons (Np/m2/s/dE) x: Energy (eV)
        alpha(float): Spectral average of incident light in cm⁻¹
        tau(float): Non-radiative recombination lifetime in s
        sigma(float): Spectral dispersion of the absorption coefficient spectrum, unitless
        dos_mass(float): Density of States effective mass in m₀
        dop_density(float): Doping density in cm⁻³
        epsilon(float): Static dielectric constant, unitless
        mu(float): Charge carrier mobility in cm²V⁻¹s⁻¹
        Tcell(float): Operating temperature of the cell in K

    Returns:
        efficiency(float): Percentage photovoltaic figure of merit efficiency relative to the SQ limit.
    '''

    Crovetto_eff = Crovetto_efficiency(E_gap, photon_spectrum, alpha, tau, sigma, dos_mass, dop_density, epsilon, mu, Tcell)
    SQ = max_eff(E_gap, photon_spectrum, Tcell)

    SQ_relative = Crovetto_eff / SQ

    return SQ_relative



import numpy as np
from scipy.integrate import simpson
import scipy.constants as sc

# q=1.60217662E-19
# kT=0.0258519975 # eV for T=300K
# k=0.000086173325 #eV/K
# h=4.135667E-15 #eVs
# c=2.9979E+8 #m/s


h = sc.h        # Planck's constant (J·s)
c = sc.c        # Speed of light (m/s)
k = sc.k        # Boltzmann constant (J/K)
q = sc.e        # Elementary charge (Coulombs)

def load_absorption(abs_file):

    abs_data = np.loadtxt(abs_file, skiprows=2)

    abs_energy_eV = abs_data[:, 0]
    abs_coeff = abs_data[:, 1]

    return abs_energy_eV, abs_coeff

def wavelength_conv(abs_energy_eV):

    wavelength_m = (h * c) / (abs_energy_eV * q)
    abs_wavelength_nm = wavelength_m * 1e9

    return abs_wavelength_nm

def extract_int_limits(E_gap):

    wavelength_min = 300

    Eg_wavelength = (( h * c ) / (E_gap * q)) * 1e9
    
    return wavelength_min, Eg_wavelength

# Make the truncated spectras

def truncate_abs_spectra(E_gap, abs_energy_eV, abs_coeff):

     wavelength_min, Eg_wavelength = extract_int_limits(E_gap)

     abs_wavelength_nm = wavelength_conv(abs_energy_eV)

     filtered_pairs = [(wl, val) for wl, val in zip(abs_wavelength_nm, abs_coeff) if wavelength_min <= wl <= Eg_wavelength]

     filtered_wavelengths_abs, filtered_abs_coff = zip(*filtered_pairs) if filtered_pairs else ([], [])

     return filtered_wavelengths_abs, filtered_abs_coff

def truncate_light_spectra(spectrum, E_gap):

     wavelength_min, Eg_wavelength = extract_int_limits(E_gap)

     spectrum = np.copy(spectrum)

     wavelength = spectrum[:, 0]
     irradiance = spectrum[:, 1]  # irradiance (W/m2/nm)

     filtered_pairs_spec = [(wl, val) for wl, val in zip(wavelength, irradiance) if wavelength_min <= wl <= Eg_wavelength]

     filtered_wavelengths_spec, filtered_irradiance_spec = zip(*filtered_pairs_spec) if filtered_pairs_spec else ([], [])

     return filtered_wavelengths_spec, filtered_irradiance_spec  

def match_wavelengths(filtered_wavelengths_abs, filtered_wavelengths_spec, filtered_irradiance_spec):
    matched_values = []

    for target_wl in filtered_wavelengths_abs:
        # Calculate absolute differences from target wavelength
        diffs = [abs(spec_wl - target_wl) for spec_wl in filtered_wavelengths_spec]
        min_diff = min(diffs)  # Find the smallest difference
        
        # Find indices where the difference is equal to min_diff
        close_indices = [i for i, d in enumerate(diffs) if d == min_diff]

        # Get the irradiance values at those indices and take the average
        closest_values = [filtered_irradiance_spec[i] for i in close_indices]
        matched_values.append(sum(closest_values) / len(closest_values))  # Average

    # matched_values = np.array(matched_values, dtype = float)

    return matched_values

def calculate_spectral_dispersion(filtered_abs_coff, matched_irradiance, filtered_wavelengths_abs):
    # Convert to log scale
    log_alpha = [np.log(a) for a in filtered_abs_coff]

    # Compute weighted mean log(α)
    numerator_mean = sum(irr * log_a for irr, log_a in zip(matched_irradiance, log_alpha)) #not mean?
    denominator_mean = sum(matched_irradiance)
    log_alpha_bar = numerator_mean / denominator_mean  # Weighted mean log(alpha)

    # Compute numerator (variance term)

    numerator_x = list(irr * (log_a - log_alpha_bar) ** 2 for irr, log_a in zip(matched_irradiance, log_alpha))
    numerator_variance = simpson(numerator_x)

    # Compute denominator (normalization term)
    denominator_variance = simpson(matched_irradiance)

    # Spectral density
    spectral_dispersion = np.sqrt(numerator_variance / denominator_variance)
    
    return spectral_dispersion

# Function to calculate spectral average
def calculate_spectral_average(filtered_abs_coff, matched_irradiance, filtered_wavelengths_abs):
    # Compute numerator (weighted sum of alpha)
    numerator_1 = list(alpha * irr for alpha, irr in zip(filtered_abs_coff, matched_irradiance))

    numerator = simpson(numerator_1)

    # Compute denominator (total irradiance)
    denominator = simpson(matched_irradiance)

    # Spectral average
    spectral_average = numerator / denominator if denominator != 0 else 0
    
    return spectral_average


def generate_spectral_parameters(abs_file, spectrum, E_gap):

    abs_energy_eV, abs_coeff = load_absorption(abs_file)
    filtered_wavelengths_abs, filtered_abs_coff = truncate_abs_spectra(E_gap, abs_energy_eV, abs_coeff)
    filtered_wavelengths_spec, filtered_irradiance_spec  = truncate_light_spectra(spectrum, E_gap)
    matched_irradiance = match_wavelengths(filtered_wavelengths_abs, filtered_wavelengths_spec, filtered_irradiance_spec)

    spectral_average = calculate_spectral_average(filtered_abs_coff, matched_irradiance, filtered_wavelengths_abs)
    spectral_dispersion = calculate_spectral_dispersion(filtered_abs_coff, matched_irradiance, filtered_wavelengths_abs)

    return spectral_average, spectral_dispersion




from setuptools import setup

setup(
    name='solphin',
    version='0.0.0',    
    description='Figure of Merit for Photovoltaics',
    url='https://github.com/SMTG-Bham/PV-FoM',
    author='Philippa U Cox, Peter P Russell',
    author_email='puc369@student.bham.ac.uk, ppr466@student.bham.ac.uk',
    license='MIT',
    packages=['solphin'],
    install_requires=['numpy',
                      'scipy',
                      'matplotlib>=3.9.4',
                      'pymatgen>=2024.8.9',
                      'mpmath',
                      'pathlib',
                      'typing_extensions',
                      'monty'                     
                      ],

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',  
        'Operating System :: POSIX :: Linux',        
        'Programming Language :: Python :: 3.9',
    ],
)